package com.persagy.dmp.rwd.controller;

import cn.hutool.core.util.StrUtil;
import com.persagy.common.criteria.JacksonCriteria;
import com.persagy.common.web.PagedResponse;
import com.persagy.dmp.config.DmpParameterStorage;
import com.persagy.dmp.rwd.config.RwdConstants;
import com.persagy.dmp.rwd.constants.CommonConstant;
import com.persagy.dmp.rwd.handler.BusinessException;
import com.persagy.dmp.rwd.handler.ResponseCode;
import com.persagy.dmp.rwd.model.RelationInstanceModel;
import com.persagy.dmp.rwd.service.BatchHandleObjectService;
import com.persagy.dmp.rwd.util.FileUtil;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.web.bind.annotation.*;
import org.springframework.web.multipart.MultipartFile;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

@RestController
@RequestMapping("/rwd/instance/object/")
@RequiredArgsConstructor
@Slf4j
public class BatchHandleObjectController {

    private final BatchHandleObjectService batchHandleObjectService;
    /***
     * Description: 下载模板接口
     * @param response : 响应对象
     * @return : void  
     * @author : lijie
     * @date :2021/9/22 9:54
     * Update By lijie 2021/9/22 9:54
     */
    @PostMapping("downloadObjectTemplateFile")
    public void downloadObjectTemplateFile(HttpServletResponse response) throws IOException {
        String groupCode = DmpParameterStorage.getGroupCode();
        if (StrUtil.isBlank(groupCode)) {
            throw new BusinessException(ResponseCode.A0402.getCode(),"缺少集团编码");
        }
        batchHandleObjectService.downloadObjectTemplateFile(response,groupCode);
    }
    /***
     * Description: 上传模板接口
     * @param file : 文件对象
     * @param request : 请求参数
     * @param response : 响应参数
     * return : void
     * @author : lijie
     * date :2021/9/24 10:10
     * Update By lijie 2021/9/24 10:10
     */
    @PostMapping("uploadObjectTemplateFile")
    public void uploadObjectTemplateFile(@RequestParam(CommonConstant.DTO_FILE) MultipartFile file,
                                          HttpServletRequest request,
                                          HttpServletResponse response){
        String groupCode = DmpParameterStorage.getGroupCode();
        if (StrUtil.isBlank(groupCode)) {
            throw new BusinessException(ResponseCode.A0402.getCode(),"缺少集团编码");
        }
        try {
            if (file.isEmpty()){
                throw new BusinessException(ResponseCode.A0400.getCode(),"文件不能为空");
            }
            String originalFilename = file.getOriginalFilename();
            if (StrUtil.isBlank(originalFilename)){
                throw new BusinessException(ResponseCode.A0400.getCode(),"文件名称不能为空");
            }
            if (!originalFilename.endsWith(CommonConstant.EXCEL_SUFFIX)){
                throw new BusinessException(ResponseCode.A0400.getCode(),"文件后缀名不符合,应该为xlsx");
            }
            batchHandleObjectService.uploadObjectTemplateFile(file,request,response);
        }catch (Exception exr){
            log.error("上传文件失败",exr);
            List<String> errorList = new ArrayList<>();
            errorList.add(exr.getMessage());
            FileUtil.writeErrorContentToResponse(errorList,response,request,CommonConstant.UPLOAD_PERSON_ERR_FILE_NAME);
        }
    }




}
